package dev.pfaff.altargui.glfw.ring;

import it.unimi.dsi.fastutil.ints.IntIterator;
import it.unimi.dsi.fastutil.longs.LongIterable;
import it.unimi.dsi.fastutil.longs.LongIterator;
import org.jspecify.annotations.NonNull;

import java.util.Arrays;
import java.util.function.LongConsumer;

public final class LongRingBuffer extends RingBuffer<Long> implements LongIterable {
	private final long[] contents;

	public LongRingBuffer(int length) {
		contents = new long[length];
	}

	public long[] getContentsInternal() {
		return contents;
	}

	@Override
	public int size() {
		return contents.length;
	}

	public long get(int i) {
		return contents[straightToRingIndex(i)];
	}

	public long getLast() {
		int i = currentIndex();
		if (i == 0) i = size() - 1;
		return contents[i];
	}

	@Override
	@Deprecated
	public void add(Long value) {
		add(value.longValue());
	}

	public void add(long value) {
		contents[inc()] = value;
	}

	/**
	 * Optimized non-overflowing average calculation.
	 */
	public double avg() {
		double avg = 0.0;

		long[] contents = this.contents;

		int len = firstRun() ? currentIndex() : contents.length;

		//noinspection UnnecessaryLocalVariable
		double dblLen = (double) len;

		// automatically vectorized by Hotspot pretty much perfectly.
		for (int i = 0; i < len; i++) {
			avg += contents[i] / dblLen;
		}

		return avg;
	}

	@NonNull
	@Override
	public LongIterator iterator() {
		return new LongIterator() {
			int i = 0;

			@Override
			public boolean hasNext() {
				return i < presentSize();
			}

			@Override
			public long nextLong() {
				assert hasNext();
				return contents[straightToRingIndex(this.i++)];
			}

			@Override
			public void forEachRemaining(LongConsumer action) {
				var size = presentSize();
				for (int i = this.i; i < size; i++) {
					action.accept(contents[straightToRingIndex(this.i)]);
				}
				this.i = size;
			}
		};
	}

	@NonNull
	public LongIterator backIterator() {
		return new LongIterator() {
			int i = presentSize();

			@Override
			public boolean hasNext() {
				return i > 0;
			}

			@Override
			public long nextLong() {
				assert hasNext();
				return contents[straightToRingIndex(--this.i)];
			}

			@Override
			public void forEachRemaining(LongConsumer action) {
				for (int i = this.i - 1; i >= 0; i--) {
					action.accept(contents[straightToRingIndex(this.i)]);
				}
				this.i = 0;
			}
		};
	}

	public IntIterator intervalCountIterator(final long from, final long to, final long interval) {
		if (interval <= 0) {
			throw new IllegalArgumentException("Argument interval must be positive, was " + interval);
		}
		return new IntIterator() {
			private final LongIterator iterator = LongRingBuffer.this.iterator();
			long t = from;

			@Override
			public boolean hasNext() {
				return t < to;
			}

			@Override
			public int nextInt() {
				assert hasNext();
				int j = 0;
				if (t == from) {
					while (iterator.hasNext()) {
						long l = iterator.nextLong();
						if (l >= t) {
							if (l < t + interval) {
								++j;
							}
							break;
						}
					}
				}
				t += interval;
				while (iterator.hasNext() && iterator.nextLong() < t) {
					++j;
				}
				return j;
			}
		};
	}

	/**
	 * Gets the count of consecutive elements that are higher than {@code thresholdValue}. If the elements are not
	 * monotonically increasing, the results are unspecified.
	 */
	public int countSince(long threshold) {
		int s = presentSize();
		if (s == 0) return 0;
		int currentIndex = currentIndex();

		// search the new region first
		int i;

		if (currentIndex != 0) {
			i = Arrays.binarySearch(contents, 0, currentIndex, threshold);

			// the value was found, return the exclusive count
			if (i >= 0) return s - 1 - ringToStraightIndex(i);

			i = -(i + 1);

			// threshold is greater than the newest value, return 0.
			if (i == currentIndex) {
				System.out.println("threshold is greater than the newest value, return 0.");
				return 0;
			}

			// threshold is between 2 elements in the new region, return the count.
			if (i > 0) return s - ringToStraightIndex(i);

			// threshold is less than the oldest value in the new region
		}

		// and there is no old region, return size
		if (currentIndex == s) return s;

		// now search the old region
		i = Arrays.binarySearch(contents, currentIndex, s, threshold);

		// the value was found, return the exclusive count
		if (i >= currentIndex) return s - 1 - ringToStraightIndex(i);

		i = -(i + 1);

		// threshold is less than the oldest value in the old region, return size
		if (i == currentIndex) return s;

		return s - 1 - ringToStraightIndex(i);
	}

	public long max(long initialValue) {
		int s = presentSize();
		long value = initialValue;
		for (int i = 0; i < s; i++) {
			value = Math.max(value, contents[straightToRingIndex(i)]);
		}
		return value;
	}

	public long min(long initialValue) {
		int s = presentSize();
		long value = initialValue;
		for (int i = 0; i < s; i++) {
			value = Math.min(value, contents[straightToRingIndex(i)]);
		}
		return value;
	}
}
